# -*- coding: utf-8 -*-

import datetime
import io
import json
import logging
import re
import time

from datetime import timedelta

from odoo import api, fields, models
from odoo.exceptions import UserError, ValidationError
from odoo.http import request
from odoo.release import version
from odoo.tools.parse_version import parse_version
from odoo.tools import groupby, str2bool

from odoo.addons.saas_worker.util import email_on_exception
from odoo.addons.saas_worker.const import TEST

try:
    from odoo.tools import pycompat
except ImportError:
    if parse_version(version) >= parse_version('11.0'):
        raise

# regex to extract 'SO2018/12344' out of ' SO2018/12344-34x2 ' in 72 steps
re_ref_split = re.compile(r"^\s*([0-9a-zA-Z/_-]+?)(?:-[0-9]+(x[0-9])?)?\s*$")
# regex to extract subscription ID 96994 out of 'SUB96994-180401_001932' in 12 steps
re_sub_split = re.compile(r"^SUB([0-9]+)(?:[^0-9][0-9_]+)?$")
# ingenico currency codes
CURRENCIES = {
    '978': 'EUR',
    '840': 'USD',
}

_logger = logging.getLogger(__name__)

class PaymentProvider(models.Model):
    _inherit = 'payment.provider'

    allow_one_shot_payments = fields.Boolean(string='Allow one shot payments', default=True)
    allow_monthly_payments = fields.Boolean(string='Allow monthly payments', default=False)
    allow_yearly_payments = fields.Boolean(string='Allow yearly payments', default=False)


    @api.model
    def _is_tokenization_required(self, sale_order_id=None, **kwargs):
        """ Allow SO for yearly subscriptions to be paid with acquirer that doesn't support tokenization """
        if sale_order_id:
            sale_order = self.env['sale.order'].browse(sale_order_id).exists()
            # yearly subscriptions can be paid using no tokenizing acquirers
            if sale_order.plan_id.billing_period_unit == 'year':
                return False

        return super()._is_tokenization_required(sale_order_id=sale_order_id, **kwargs)

    @api.model
    def _get_compatible_providers(self, company_id, *args, sale_order_id=None, **kwargs):
        # Use `sale_order_id` to automatically filter appropriate provider for
        # subscription accordingly to it's recurrence.
        filter_payment_frequency = None
        if sale_order_id and isinstance(sale_order_id, str):
            # `/my/payment_manage` do pass sale_order_id as string
            sale_order_id = int(sale_order_id)
        order_sudo = self.env["sale.order"].browse(sale_order_id).sudo() if sale_order_id else self.env["sale.order"]
        if order_sudo and order_sudo.is_subscription:
            filter_payment_frequency = order_sudo.plan_id.billing_period_unit
            if filter_payment_frequency not in ("month", "year"):
                filter_payment_frequency = "one_shot"

            # OPW-3827812: when managing subscription (/my/payment_method)
            # always display providers of the subscription's company
            manage_subscription = kwargs.get("manage_subscription")
            if isinstance(manage_subscription, str):
                manage_subscription = str2bool(manage_subscription)
            if manage_subscription:
                company_id = order_sudo.company_id.id

        if filter_payment_frequency == "month":
            # for monthly subscription, we want to enforce use of token for reccuring invoicing.
            kwargs["force_tokenization"] = True

        providers = super()._get_compatible_providers(company_id, *args, sale_order_id=sale_order_id, **kwargs)

        # We add razor pay if the company is india even if it doesn't tokenize
        if company_id == self.env.ref('openerp_enterprise.company_in').id:
            providers |= self.env.ref('openerp_enterprise.payment_acquirer_razorpay_in')
        # Add Stripe HK for Odoo ID, until they have a local payment provider
        elif company_id == self.env.ref('openerp_enterprise.company_idn').id:
            providers |= self.env.ref('openerp_enterprise.payment_acquirer_stripe_hk')
        providers = providers.sorted('sequence')

        # filter out the sepa direct debit providers and allow only authorized partners to use it
        providers = providers._oe_filter_out_sepa_partners()
        # ensure we only use test providers on TEST and production providers on prod.
        providers = providers.filtered(lambda ac: ac.state == ("test" if TEST else "enabled"))

        if filter_payment_frequency:
            # filter only allowed recurrency
            if filter_payment_frequency == "month":
                allowed_payment_frequency = "allow_monthly_payments"
            elif filter_payment_frequency == "year":
                allowed_payment_frequency = "allow_yearly_payments"
            else:
                allowed_payment_frequency = "allow_one_shot_payments"
            providers = providers.filtered(allowed_payment_frequency)

        # fallback to Ogone for specific companies having no available provider
        # (except for "validation" operation, as it's not supported by Ogone)
        if not providers and not kwargs.get('is_validation'):
            if company_id in (
                self.env.ref('openerp_enterprise.company_dmcc').id,
                self.env.ref('openerp_enterprise.company_ke').id,
            ):
                providers = self.env.ref("openerp_enterprise.payment_acquirer_ogone")
        return providers

    def _oe_filter_out_sepa_partners(self):
        """ Method to filter out the sepa direct debit payment acquirers for partners that are not
            allowed to use it, until we are ready to use it massively.
            The partner ids are retrieved from the `ir.config_parameter` => `oe_sepa_allowed_partner_ids`.
            Those ids are partner commercial ids, thus allowing all users under a company to pay using sepa.
        """
        # check if the current user is allowed, do it only for payments that are done through a http request
        # this way, we do not block anything running in a cron
        if request:
            # retrieve the partner ids
            allowed_partner_ids = self.env['ir.config_parameter'].sudo().get_param('oe_sepa_allowed_partner_ids', '[]')
            # json load them
            try:
                allowed_partner_ids = json.loads(allowed_partner_ids)
            except Exception:
                _logger.error("Invalid ir.config_parameter oe_sepa_allowed_partner_ids: %r", allowed_partner_ids)
                allowed_partner_ids = []

            user = self.env['res.users'].sudo().browse(request.session.uid)
            commercial_partner_id = user.commercial_partner_id
            if not commercial_partner_id or commercial_partner_id.id not in allowed_partner_ids:
                return self.filtered(lambda x: x.code != 'sepa_direct_debit')
        return self


    ###################################
    # SPECIFIC OGONE PAYMENT MANAGEMENT
    ###################################

    def _locate_partner_by_tx_ref(self, ref):
        """try to locate a partner via a matching Sales Order, Invoice or Subscription

        Supported tx ref formats:
        - SO2021/1476408-20210802132510
        - SO2021/1476408-34x2
        - SO2021/1476408
        - SUB243151-210519_012147
        - SUB227543-20210715130923
        - SUB243151
        """
        self.ensure_one()
        partner = None
        match = re_ref_split.findall(ref)
        if match:
            # keep only 'SO2018/12344' out of ' SO2018/12344-34x2 '
            ref = match[0][0]
        if ref.startswith('SO'):
            ref = ref.replace('-', '/')
            orders = self.env['sale.order'].search(
                [('name', '=', ref.replace('-', '/')), ('company_id', '=', self.company_id.id)],
                limit=1)
            _logger.info('Payment ref %s -> SO match: %s', ref, orders)
            if orders:
                partner = orders[0].partner_id
        elif ref.startswith('SUB'):
            # locate matching subscription ID 96994 out of 'SUB96994-YYMMDD_HHMMSS' or 'SUB96994'
            match = re_sub_split.findall(ref)
            if match:
                subscription = self.env['sale.subscription'].browse(int(match[0])).exists() # TBEB TODO: not sure the matching of sub id will work like this, need to check
                _logger.info('Payment ref %s -> SUB match: %s', ref, subscription)
                partner = subscription.partner_id
        else:
            # fallback search in invoices
            invoices = self.env['account.move'].search(
                [('name', '=', ref), ('company_id', '=', self.company_id.id)],
                limit=1)
            _logger.info('Payment ref %s -> Invoices fallback: %s', ref, invoices)
            if invoices:
                partner = invoices[0].partner_id
        _logger.info('Payment ref %s -> Partner located: %s', ref, partner)
        return partner

    def _ogone_import_payout_csv(self, csv_bytes):
        reader = pycompat.csv_reader(
            io.BytesIO(csv_bytes),
            quotechar='"',
            delimiter=';')
        batch = []
        columns = None
        for row in reader:
            if row[0] == 'HEADER':
                columns = row
                continue
            if columns and row[0] == 'MATCHING':
                # all payment rows start with 'MATCHING'
                payment = dict(zip(columns, row))
                batch.append(payment)
            else:
                # skip other rows (e.g. TITLE or END lines)
                continue
        if batch:
            return self._ogone_import_payout_batch(batch)

    def _ogone_generate_signature(self, values, incoming=True, format_keys=False):
        # To avoid copy-pasting ogone's `_get_specific_rendering_values()` and
        # `_send_payment_request()` to add specific 3DS and COF parameters, we
        # are using the fact that `values` is passed by reference to update the
        # source dict before actually computing the signature
        if not incoming:
            if values.get('PM', '') == 'BCMC':
                # 'BCMC' is invalid as payment method value; we have to use
                # 'CreditCard' and filter for 'BCMC' brand instead, so that user
                # immediately land on the dedicated bancontact payment page.
                values.update({'PM': 'CreditCard', 'BRAND': 'BCMC'})
            else:
                # or when user chosen another payment method, always exclude
                # Bancontact/MisterCash (noting that this PSP can't be used
                # later on for recurring payment)
                values["EXCLPMLIST"] = "BCMC"

        if not incoming and 'ALIAS' in values:
            # initial payment request w/ tokenization
            # called from `_get_specific_rendering_values()`
            # (see payment_ogone/models/payment_transaction.py)
            if 'ALIASUSAGE' in values:
                values.update({
                    # add specific parameters to force strong 3DS when doing
                    # an initial payment the intent to create a new payment token
                    "MPI.THREEDSREQUESTORCHALLENGEINDICATOR": "04",  # 04 = Mandate

                    # add COF parameters to increase the chance of transaction
                    # being authorized by the bank
                    "COF_INITIATOR": "CIT",
                    "COF_SCHEDULE": "UNSCHED",
                    "COF_TRANSACTION": "FIRST",
                })
            # subsequent payment requests w/ token (recurring)
            # called from `_send_payment_request()`
            # (see payment_ogone/models/payment_transaction.py)
            elif values.get('ECI') == 9:
                values.update({
                    # add COF parameters to increase the chance of transaction
                    # being authorized by the bank
                    "COF_INITIATOR": "MIT",
                    "COF_SCHEDULE": "UNSCHED",
                    "COF_TRANSACTION": "SUBSEQ",
                })
        shasign = super()._ogone_generate_signature(values, incoming=incoming, format_keys=format_keys)
        if "MPI.THREEDSREQUESTORCHALLENGEINDICATOR" in values:
            # make it suitable for qweb template
            values["MPI_THREEDSREQUESTORCHALLENGEINDICATOR"] = values.pop("MPI.THREEDSREQUESTORCHALLENGEINDICATOR")
        return shasign

    @email_on_exception(to="sys+ogone@odoo.com")
    def _ogone_import_payout_batch(self, payment_data):
        """ Process an Ingenico payout batch, as provided by standard Ingenico Payout push reports,
        and passed as a list of dicts with this format:

            [
               {
                 'TRANSACTION_ID':    '4004918293',
                 'PAYMENT_DATE':      '20180407',
                 'CURRENCY_CODE':     '978',     # 840=USD, 978=EUR  (or TRANSACTION_CURRENCY)
                 'CARD_TYPE':         'VISA',    # VISA, Mastercard, ...
                 'ORDER_ID':          'SUB64739-180406_000653',
                 'PAYOUT_DATE':       '20180420',
                 'REMITTANCE_NB':     '22240652',# or PAYOUT_REFERENCE
                 'BRUT_AMOUNT':       '15000',   # Gross Amount * 100  (or GROSS_AMOUNT)
                 'NET_AMOUNT':        '14600',   # Net Amount * 100
                 'COMMISSION_AMOUNT': '400',     # Comm Amount * 100
                 'PAYOUT_ACCOUNT':    'BE123456',# IBAN number
                },
               ...
            ]

        Payments will be grouped by "REMITTANCE_NB", i.e payout report, and imported as batch
        payments, if that batch is not yet imported.

        See also the reference for payout reports:
        https://payment-services.ingenico.com/int/en/ogone/support/guides/user%20guides/push-reports/create-new-report#createreportreconciliation
        """
        Payment = self.env['account.payment']
        BatchDep = self.env['account.batch.payment']
        Currency = self.env['res.currency']
        currencies = dict((c.name.upper(), c.id) for c in Currency.search([]))
        ogone_pm_line = self.env.ref('openerp_enterprise.payment_method_line_ogone')
        ogone_pm = ogone_pm_line.payment_method_id

        # At the beginning of the month, some payments may be imported with a date
        # in the previous month, but the lock date may have already moved.
        # In that case we should move them to the first day after the lock date (first day
        # of the next month)
        lock_date = self.company_id._get_user_fiscal_lock_date()

        def get_payment_date(payment_date):
            return max(payment_date, lock_date + timedelta(days=1))

        def fix_payment_method(payment):
            if payment.payment_method_line_id != ogone_pm_line:
                payment.payment_method_line_id = ogone_pm_line.id
            # sometimes there's a mismatch between payment method and payment method line,
            # but we need both to be forced so we can validate the batch
            if payment.payment_method_id != ogone_pm:
                payment.payment_method_id = ogone_pm.id

        # Group payments by payout number, to obtain batches equivalent to Payouts
        def by_payout(p):
            return p.get('REMITTANCE_NB') or p.get('PAYOUT_REFERENCE')
        grouped_tx = groupby(payment_data, by_payout)

        # process batches
        for key, payment_batch in grouped_tx:
            row = payment_batch[0]
            day = datetime.datetime.strptime(row.get('PAYOUT_DATE'), '%Y%m%d')
            cur_code = row.get('CURRENCY_CODE') or row.get('TRANSACTION_CURRENCY')
            cur_name = CURRENCIES.get(cur_code, cur_code)
            currency_id = currencies.get(cur_name)
            _logger.info(
                'Ingenico: process payout %s with %s transactions in %s',
                key, len(payment_batch), cur_code)
            batch_name = 'Ingenico_Payout_%s_%s' % (key, cur_name)
            batch_domain = [('name', '=', batch_name), ('journal_id', '=', self.journal_id.id)]
            if BatchDep.search(batch_domain):
                _logger.info('Ingenico batch %s already imported, skipping.', batch_name)
                continue

            # Create batch payment immediately to prevent concurrent creation (dupe emails)
            batch_payment = BatchDep.create({
                'name': batch_name,
                'date': day,
                'journal_id': self.journal_id.id,
                'batch_type': 'inbound',
            })
            self.env.cr.commit()

            batch_payments = []
            for payment in payment_batch:
                ref = payment.get('ORDER_ID')
                pay_id = payment.get('TRANSACTION_ID')
                payment_type = ('inbound'
                                if payment.get('OPERATION_TYPE') in ('DT', 'TDEB')
                                else 'outbound')
                amount_str = payment.get('BRUT_AMOUNT') or payment.get('GROSS_AMOUNT')
                if not amount_str:
                    # skip fee/info line, probably from Payout_with_fees report format
                    continue
                amount = float(amount_str) / 100.0

                # locate payment based on transaction
                tx_domain = [('provider_reference', '=', pay_id), ('provider_id', '=', self.id)]
                tx = self.env['payment.transaction'].search(tx_domain, limit=1)
                _logger.info('Ingenico payment %s (PAYID %s) -> tx match: %s', ref, pay_id, tx)

                if tx:
                    tx_payment = Payment.search([
                        ('payment_transaction_id', '=', tx.id),
                        ('journal_id', '=', self.journal_id.id),
                        ('payment_type', '=', payment_type)], limit=1)
                    if tx_payment:
                        _logger.info('Ingenico payment %s (PAYID %s) found by tx', ref, pay_id)
                        batch_payments.append((4, tx_payment.id))
                        fix_payment_method(tx_payment)
                        continue
                    # no payment linked, search via invoice
                    tx_payment = tx.invoice_ids._get_reconciled_payments() \
                                               .filtered(lambda x: x.payment_type == payment_type and
                                                                   x.journal_id == self.journal_id)
                    if tx_payment:
                        tx_payment = tx_payment[0]
                        _logger.info('Ingenico payment %s (PAYID %s) found via invoices %s',
                                     ref, pay_id, tx.invoice_ids.ids)
                        batch_payments.append((4, tx_payment.id))
                        fix_payment_method(tx_payment)
                        continue

                # otherwise attempt to match by communication/reference
                payment_domain = [
                    ('payment_type', '=', payment_type),
                    ('partner_type', '=', 'customer'),
                    ('currency_id', '=', currency_id),
                    ('journal_id', '=', self.journal_id.id),
                    # ('payment_date', '=', day), # Payment date often 1d later
                    '|', '|', ('payment_reference', '=', ref), ('ref', '=', ref), ('ref', '=', "[MISSING] %s" % ref),
                    ('batch_payment_id', '=', False)]
                tx_payment = Payment.search(payment_domain, limit=1)

                # link valid payments but ignore null payments that will be rejected in the batch,
                # probably caused by a manual input mistake
                if tx_payment and tx_payment.amount:
                    _logger.info('Ingenico payment %s (PAYID %s) found by communication', ref, pay_id)
                    batch_payments.append((4, tx_payment.id))
                    fix_payment_method(tx_payment)
                    continue

                # no match! search for partner then create the missing payment
                partner = tx.partner_id or tx.invoice_ids.mapped('partner_id') or tx.sale_order_ids.mapped('partner_id')
                if not partner:
                    partner = self._locate_partner_by_tx_ref(tx.reference or ref)

                # prepare the payment
                payment_date = payment.get('PAYMENT_DATE')
                if payment_date:
                    # FIXME: does format depend on Ogone account language?
                    payment_date = datetime.datetime.strptime(payment_date, '%Y%m%d').date()

                attempts = 0
                self.env.cr.commit() # prepare for new payment loop
                while attempts < 10:
                    try:
                        attempts += 1
                        payment_vals = {
                            'payment_type': payment_type,
                            'partner_type': 'customer',
                            'currency_id': currency_id,
                            'journal_id': self.journal_id.id,
                            'payment_reference': "[MISSING] %s" % ref,
                            'ref': "[MISSING] %s" % ref,
                            'amount': amount,
                            'payment_transaction_id': tx and tx.id,
                            'partner_id': partner and partner[0].id,
                            'date': get_payment_date(payment_date),  # We don't have Ogone's "order date" handy
                            'payment_method_line_id': ogone_pm_line.id,
                            'payment_method_id': ogone_pm.id,
                        }
                        tx_payment = Payment.create(payment_vals)
                        tx_payment.action_post()

                        # force a new tx, avoid a super-long running transaction that is more likely to be interrupted
                        self.env.cr.commit()
                        break
                    except Exception as e:
                        _logger.info('Exception posting payment: %r, try %s/10', e, attempts)
                        self.env.cr.rollback()
                        time.sleep(1)
                        continue
                    raise Exception("Failed posting payment after 10 attempts")

                _logger.info('Ingenico payment %s (PAYID %s) was MISSING - created as %s', ref, pay_id, tx_payment.name)
                batch_payments.append((4, tx_payment.id))

            _logger.info('Ingenico batch import %s with %s payments', batch_name, len(batch_payments))
            try:
                batch_payment.write({
                    'payment_ids': batch_payments
                })
                self.env.cr.commit()
            except (UserError, ValidationError):
                # auto-posting could fail due to lock date or presence of a zero-amount payment
                # don't crash as it could be a multi-batch and the next one still needs to be imported.
                _logger.exception('Failed to post batch payment %s, may need manual review', batch_name)
            except Exception:
                # make sure the email exception can be sent even if tx is dead
                _logger.exception('Batch payment crashed, rolling back: %s', batch_name)
                self.env.cr.rollback()
                raise


class PaymentMethod(models.Model):
    _inherit = "payment.method"

    @api.model
    def _get_compatible_payment_methods(self, *args, sale_order_id=None, **kwargs):
        if sale_order_id:
            sale_order = self.env["sale.order"].browse(sale_order_id)
            if sale_order.is_subscription or sale_order.subscription_id.is_subscription:
                # since odoo/enterprise@ab815abb1, Odoo enforce tokenization
                # support for payment methods of a subscription (de facto
                # excluding "Wire Payment"). As we still need to allow
                # non-tokenized payment methods on (for ex. Yearly or
                # Partnership) subscriptions, set an empty sale_order_id
                # to avoid that behavior.
                pm_kwargs = dict(kwargs, sale_order_id=None)
                payment_methods = super()._get_compatible_payment_methods(*args, **pm_kwargs)
                if sale_order.is_subscription and sale_order.plan_id.billing_period_unit == 'month':
                    payment_methods = payment_methods.filtered(
                        lambda pm: pm.support_tokenization or pm.code == 'bancontact'
                    )
                return payment_methods
        return super()._get_compatible_payment_methods(*args, sale_order_id=sale_order_id, **kwargs)


# Override account.payment to support Payout import via email
class AccountPayment(models.Model):
    _inherit = "account.payment"

    @api.model
    def message_new(self, msg_dict, custom_values=None):
        custom_values = custom_values or {}
        attachments = msg_dict.get('attachments')
        subject = msg_dict.get('subject', '').lower()
        provider_id = custom_values.get('provider_id') or custom_values.get('acquirer_id')

        if not (attachments and 'payout' in subject and provider_id):
            return super(AccountPayment, self).message_new(msg_dict, custom_values)

        provider = self.env['payment.provider'].browse(provider_id)
        assert provider.code == 'ogone'

        for attachment in attachments:
            try:
                content = attachment.content
                if isinstance(content, str):
                    content = content.encode()
                provider._ogone_import_payout_csv(content)
            except Exception:
                _logger.exception('Failed to parse payout batch via email')
        self.env.cr.commit()
        # disable message_post, creation is a side-effect
        return self.env['mail.thread']
